/*
 * Copyright (c) 2009 The Trustees of Indiana University and Indiana
 *                    University Research and Technology
 *                    Corporation.  All rights reserved.
 *
 * Author(s): Torsten Hoefler <htor@illinois.edu>
 *
 */

#include "netgauge.h"
#ifdef NG_PTRN_MEMORY
#include "hrtimer/hrtimer.h"
#include "ptrn_memory_cmdline.h"
#include "ng_tools.hpp"
#include <vector>
#include <time.h>
#include <algorithm>
#include <numeric>


extern "C" {

// the type for the basic memory access
#define TYPE double

/* this only exists to prevent compiler optimizations that remove the
 * cachewiper ;) */
volatile TYPE NG_Memory_res=0;


extern struct ng_options g_options;

/* internal function prototypes */
static void memory_do_benchmarks(struct ng_module *module);

/**
 * comm. pattern description and function pointer table
 */
static struct ng_comm_pattern pattern_memory = {
   pattern_memory.name = "memory",
   pattern_memory.desc = "measures memory performance",
   pattern_memory.flags = 0,
   pattern_memory.do_benchmarks = memory_do_benchmarks
};

/**
 * register this comm. pattern for usage in main
 * program
 */
int register_pattern_memory() {
   ng_register_pattern(&pattern_memory);
   return 0;
}


/* both fucntions come from ptrn_noise.c and is a bit nasty, but I don't
 * want to put it in netgauge.c because it might not be portable :-( */
void write_host_information(FILE *fd);
FILE *open_output_file(char *filename);

// this is a *function* with a *external volatile side-effect* so that it can't be moved around
TYPE wipe_cache(TYPE *cachewiper, int cachewiper_size) {
  TYPE ret=0;
  for(int w=0; w<cachewiper_size; w++) {
    ret += ++(cachewiper[w]);
  }
  NG_Memory_res += ret;
  return ret;
}

static void memory_do_benchmarks(struct ng_module *module) {

  /** currently tested packet size and maximum */
  long data_elems;

  /** number of times to test the current datasize */
  long test_count = g_options.testcount;

  /** counts up to test_count */
  int test_round = 0;

  //parse cmdline arguments
  struct ptrn_memory_cmd_struct args_info;
  //printf("The string I got: %s\n", g_options.ptrnopts);
  if (ptrn_memory_parser_string(g_options.ptrnopts, &args_info, "netgauge") != 0) {
    exit(EXIT_FAILURE);
  }

  long max_data_elems = 33554432/sizeof(TYPE);  // 32 MiB constant!  ng_min(g_options.max_datasize + module->headerlen, module->max_datasize);
  /*Allocating buffers for the source and destination*/
  /* get needed data buffer memory */
  ng_info(NG_VLEV1, "Allocating %d bytes for source data buffer", max_data_elems*sizeof(TYPE));
  TYPE *buffer1; 
  NG_MALLOC(module, TYPE*, max_data_elems*sizeof(TYPE), buffer1);

  ng_info(NG_VLEV2, "Initializing source buffer (make sure it's really allocated)");
  for (int i = 0; i < max_data_elems; i++) buffer1[i] = (TYPE)0xff;

 /* get needed data buffer memory */
  ng_info(NG_VLEV1, "Allocating %d bytes for destination data buffer", max_data_elems*sizeof(TYPE));
  TYPE *buffer2; 
  NG_MALLOC(module, TYPE*, max_data_elems*sizeof(TYPE), buffer2);

  ng_info(NG_VLEV2, "Initializing destination buffer (make sure it's really allocated)");
  for (int i = 0; i < max_data_elems; i++) buffer2[i] = (TYPE)0xff;

 /* get needed data buffer memory */
  static const int cachewiper_size = 1048576*40/sizeof(TYPE); // 40 MiB should be enough for now
  ng_info(NG_VLEV1, "Allocating %d bytes for cache wipe buffer", cachewiper_size*sizeof(TYPE));
  TYPE *cachewiper; 
  NG_MALLOC(module, TYPE*, cachewiper_size*sizeof(TYPE), cachewiper);

  ng_info(NG_VNORM, "****** ATTENTION: Ensure that 'ptrn_memory.cpp' was compiled with >= -O3! ******");
  if(args_info.wipe_given) ng_info(NG_VNORM, "Wiping cache with %.2f MiB before each measurement", (double)cachewiper_size*sizeof(TYPE)/1024/1024);
  if(args_info.memcpy_given) ng_info(NG_VNORM, "Using memcpy() instead of loop for memory copy");

  int rank = g_options.mpi_opts->worldrank;
  
  /* buffer for header ... */
  char* txtbuf = (char *)malloc(2048 * sizeof(char));
  if (txtbuf == NULL) {
    ng_error("Could not (re)allocate 2048 byte for output buffer");
    ng_exit(10);
  }
  memset(txtbuf, '\0', 2048);
  
  char fname[1024];
  strncpy(fname, g_options.output_file, 1023);
  if(args_info.write_all_given) { 
    char suffix[512];
    snprintf(suffix, 511, ".%i", g_options.mpi_opts->worldrank);
    strncat(fname, suffix, 1023);
  }
	FILE* outputfd = open_output_file(fname);
  write_host_information(outputfd);


  /* header printing */
  snprintf(txtbuf, 2047,
          "## Netgauge v%s - mode %s - 2 processes\n"
          "##\n"
    "## A...message size [byte]\n"
    "##\n"
    "## B...minimum time\n"
    "## C...average time\n"
    "## D...median time\n"
    "## E...maximum time\n"
    "## F...standard deviation (stddev)\n"
    "## G...number of measurements, that were bigger than avg + 2 * stddev.\n"
    "##\n"
    "## three blocks: - [read] - [write] - [copy] \n"
    "##\n"
    "## A  -  B  C  D  E (F G) - B  C  D  E (F G) -  B  C  D  E (F G)\n",
    NG_VERSION,
    g_options.mode);


  if(!rank || args_info.write_all_given) {
    fprintf(outputfd, "%s", txtbuf);
    fprintf(outputfd, "#\n"
        "# gnuplot script (medians):\n"
        "#  set logscale x\n"
        "#  plot 'ng.out' using 1:($1/$5) title 'read'\n"
        "#  replot 'ng.out' using 1:($1/$12) title 'write'\n"
        "#  replot 'ng.out' using 1:($1/$19) title 'copy'\n"
        "# \n"
    );
  }

  if(rank == 0) {
    // if very verbose - long output
    if (NG_VLEV2 & g_options.verbose) {
      printf("%s", txtbuf);
    } else
    // if verbose - short output
    if (NG_VLEV1 & g_options.verbose) {
      snprintf(txtbuf, 2047,
              "## Netgauge v%s - mode %s - 2 processes\n"
              "##\n"
        "## A...message size [byte]\n"
        "##\n"
        "## B...minimum time\n"
        "## C...average time\n"
        "## D...median time\n"
        "## E...maximum time\n"
        "##\n"
        "## three blocks: - [read] - [write] - [copy] \n"
        "##\n"
        "## A  -  B  C  D  E - B  C  D  E -  B  C  D  E \n",
        NG_VERSION,
        g_options.mode);


      printf("%s", txtbuf);
      
    } else
    // if not verbose - short output
    {
      // no header ...
    }
  }

  /* Outer test loop
   * - geometrically increments data_elems (i.e. data_elems = data_elems * 2)
   *  (- geometrically decrements test_count) not yet implemented
   */
  for (data_elems = g_options.min_datasize; data_elems <= max_data_elems; data_elems*=2) {
    ++test_round;

    // the benchmark results
    std::vector<double> tr, tw, tc; // read, write, copy
    
    ng_info(NG_VLEV1, "Round %d: testing %d times with %d bytes:", test_round, test_count, data_elems);
    // if we print dots ...
    if ( (rank==0) && (NG_VLEV1 & g_options.verbose) ) {
      printf("# ");
    }

    /* Inner test loop
     * - run the requested number of tests for the current data size
     * - but only if testtime does not exceed the max. allowed test time
     *   (only if max. test time is not zero)
     */
    for (int test = -1 /* 1 warmup test */; test < test_count; test++) {
	
	    /* first statement to prevent floating exception */
      /* TODO: add cool abstract dot interface ;) */
	    if ( (NG_VLEV1 & g_options.verbose) && ( test_count < NG_DOT_COUNT || !(test % (int)(test_count / NG_DOT_COUNT)) )) {
	      printf(".");
	      fflush(stdout);
	    }

      /* do the client stuff ... take time, send message, wait for
       * reply and take time  ... simple ping-pong scheme */
      HRT_TIMESTAMP_T t[3];  
      unsigned long long tirtt;
  
      ////////////////////////////////////////////////////////////
      // read phase
      if(args_info.wipe_given) NG_Memory_res += wipe_cache(cachewiper, cachewiper_size);

      MPI_Barrier(MPI_COMM_WORLD);      

      HRT_GET_TIMESTAMP(t[0]);
  
      {
        register TYPE k=0;
        for(register int i=0; i<data_elems; ++i) {
          k+=buffer1[i];
        }
        NG_Memory_res += k; // avoid optimization
      }

      HRT_GET_TIMESTAMP(t[2]);
      HRT_GET_ELAPSED_TICKS(t[0],t[2],&tirtt);
  
      if(test >= 0) tr.push_back(HRT_GET_USEC(tirtt));
      
      ////////////////////////////////////////////////////////////
      // write phase
      if(args_info.wipe_given) NG_Memory_res += wipe_cache(cachewiper, cachewiper_size);

      MPI_Barrier(MPI_COMM_WORLD);      

      HRT_GET_TIMESTAMP(t[0]);
  
      {
        register TYPE k=0;
        for(register int i=0; i<data_elems; ++i) {
          buffer1[i]=k++;
        }
        NG_Memory_res += k; // avoid optimization
      }

      HRT_GET_TIMESTAMP(t[2]);
      HRT_GET_ELAPSED_TICKS(t[0],t[2],&tirtt);
  
      if(test >= 0) tw.push_back(HRT_GET_USEC(tirtt));
      
      ////////////////////////////////////////////////////////////
      // copy phase
      if(args_info.wipe_given) NG_Memory_res += wipe_cache(cachewiper, cachewiper_size);

      MPI_Barrier(MPI_COMM_WORLD);      

      if(args_info.memcpy_given) {
        HRT_GET_TIMESTAMP(t[0]);
        {
          memcpy(buffer2, buffer1, data_elems*sizeof(TYPE));
        }
        HRT_GET_TIMESTAMP(t[2]);
      } else { 
        HRT_GET_TIMESTAMP(t[0]);
        {
          register TYPE k=0;
          for(register int i=0; i<data_elems; ++i) {
            buffer1[i]=buffer2[i];
          }
          NG_Memory_res += k; // avoid optimization
        }
      }

      HRT_GET_TIMESTAMP(t[2]);
      HRT_GET_ELAPSED_TICKS(t[0],t[2],&tirtt);
  
      if(test >= 0) tc.push_back(HRT_GET_USEC(tirtt));


    }	/* end inner test loop */
    
    /* compute statistics - read time */
    double tr_avg = std::accumulate(tr.begin(), tr.end(), (double)0)/(double)tr.size(); 
    double tr_min = *min_element(tr.begin(), tr.end()); 
    double tr_max = *max_element(tr.begin(), tr.end()); 
    std::vector<double>::iterator nthr = tr.begin()+tr.size()/2;
    std::nth_element(tr.begin(), nthr, tr.end());
    double tr_med = *nthr;
    double tr_var = standard_deviation(tr.begin(), tr.end(), tr_avg);
    int tr_fail = count_range(tr.begin(), tr.end(), tr_avg-tr_var*2, tr_avg+tr_var*2);

    /* compute statistics - write time */
    double tw_avg = std::accumulate(tw.begin(), tw.end(), (double)0)/(double)tw.size(); 
    double tw_min = *min_element(tw.begin(), tw.end()); 
    double tw_max = *max_element(tw.begin(), tw.end()); 
    std::vector<double>::iterator nthw = tw.begin()+tw.size()/2;
    std::nth_element(tw.begin(), nthw, tw.end());
    double tw_med = *nthw;
    double tw_var = standard_deviation(tw.begin(), tw.end(), tw_avg);
    int tw_fail = count_range(tw.begin(), tw.end(), tw_avg-tw_var*2, tw_avg+tw_var*2);

    /* compute statistics - copy time */
    double tc_avg = std::accumulate(tc.begin(), tc.end(), (double)0)/(double)tc.size(); 
    double tc_min = *min_element(tc.begin(), tc.end()); 
    double tc_max = *max_element(tc.begin(), tc.end()); 
    std::vector<double>::iterator nthc = tc.begin()+tc.size()/2;
    std::nth_element(tc.begin(), nthc, tc.end());
    double tc_med = *nthc;
    double tc_var = standard_deviation(tc.begin(), tc.end(), tc_avg);
    int tc_fail = count_range(tc.begin(), tc.end(), tc_avg-tc_var*2, tc_avg+tc_var*2);


    // if very verbose - long output
      memset(txtbuf, '\0', 2048);
      snprintf(txtbuf, 2047,
	      "%ld -  %.2lf %.2lf %.2lf %.2lf (%.2lf %i) -  %.2lf %.2lf %.2lf %.2lf (%.2lf %i) -  %.2lf %.2lf %.2lf %.2lf (%.2lf %i)\n",
        data_elems*sizeof(TYPE), 
        tr_min, tr_avg, tr_med, tr_max, 
        tr_var, /* standard deviation */
        tr_fail, /* how many are bigger than twice the standard deviation? */
        tw_min, tw_avg, tw_med, tw_max, tw_var, tw_fail, 
        tc_min, tc_avg, tc_med, tc_max, tc_var, tc_fail);
    if(!rank || args_info.write_all_given) fprintf(outputfd, "%s", txtbuf);


	  if (rank==0) {
      /* add linebreak if we made dots ... */
      if ( (NG_VLEV1 & g_options.verbose) ) {
        ng_info(NG_VLEV1, "\n");
      }

      if (NG_VLEV2 & g_options.verbose) {
        printf("%s", txtbuf);
        
      } else
      // if verbose - short output
      if (NG_VLEV1 & g_options.verbose) {
        memset(txtbuf, '\0', 2048);
        snprintf(txtbuf, 2047,
	        "%ld - %.2lf %.2lf %.2lf %.2lf - %.2lf %.2lf %.2lf %.2lf - %.2lf %.2lf %.2lf %.2lf\n",
          data_elems*sizeof(TYPE), 
          
          tr_min, tr_avg, tr_med, tr_max, 
          tw_min, tw_avg, tw_med, tw_max, 
          tc_min, tc_avg, tc_med, tc_max);
        printf("%s", txtbuf);
        
      } else
      // if not verbose - short output
      {
        memset(txtbuf, '\0', 2048);
        snprintf(txtbuf, 2047,
	        "%ld bytes \t r w c -> %.2lf %.2lf %.2lf us \t\t ( %.2lf %.2lf %.2lf MiB/s)\n",
          data_elems*sizeof(TYPE), 
          tr_avg, tw_avg, tc_avg, 
          (double)data_elems*sizeof(TYPE)/tr_avg,
          (double)data_elems*sizeof(TYPE)/tw_avg,
          (double)data_elems*sizeof(TYPE)/tc_avg);
        printf("%s", txtbuf);
        
      }
    }

    ng_info(NG_VLEV1, "\n");
    fflush(stdout);

  }	/* end outer test loop */

  if(!rank || args_info.write_all_given) fclose(outputfd);


  for(int w=0; w<cachewiper_size; w++) {
    NG_Memory_res += cachewiper[w];
  }

   
 shutdown:
   if(txtbuf) free(txtbuf);
    
}

} /* extern C */

#else
int register_pattern_memory(void) {return 0;};
#endif
